#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"

#include <fstream>
using namespace std;

TeamBox TBox;

TeamEntry *TeamBox::FindTeam(Uint16 Freq)
{
	_listnode <TeamEntry> *parse = Box.head;
	while (parse)
	{
		if (parse->item->Freq == Freq)
			return parse->item;

		parse = parse->next;
	}

	return NULL;
}

TeamEntry *TeamBox::FindTeam(char *Owner)
{
	_listnode <TeamEntry> *parse = Box.head;
	while (parse)
	{
		if (CMPSTR(parse->item->Owner, Owner))
			return parse->item;

		parse = parse->next;
	}

	return NULL;
}

bool TeamBox::LoadFromDisk()
{
	ifstream file(VEX_TEAMFILE, ios::binary);
	if (file)
	{
		file.seekg(0, ios::end);

		if (file.tellg() % sizeof(TeamEntry))
			return false;

		file.seekg(0, ios::beg);

		while (!file.eof())
		{
			TeamEntry *t = new TeamEntry();
			char *c = new char[sizeof(TeamEntry)];

			file.read(c, sizeof(TeamEntry));

			DoXOR(c, sizeof(TeamEntry));
			memcpy(t, c, sizeof(TeamEntry));

			Box.append(t);

			delete []c;
		}
	}

	return true;
}

bool TeamBox::SaveToDisk()
{
	ofstream file(VEX_TEAMFILE, ios::binary);

	if (file)
	{
		_listnode <TeamEntry> *parse = Box.head;
		while (parse)
		{
			char c[sizeof(TeamEntry)];
			memcpy(c, parse->item, sizeof(TeamEntry));

			DoXOR(c, sizeof(TeamEntry));

			file.write(c, sizeof(TeamEntry));

			parse = parse->next;
		}
	}

	return true;
}

inline void TeamBox::DoXOR(char *buffer, Uint32 len)
{
	for (Uint32 i = 0; i < len; i += 4)
		*(Uint32*)&buffer[i] ^= VEX_XORKEY;
}

TeamEntry *TeamBox::AddTeam(Uint16 Freq, char *Owner, Uint32 Password)
{
	TeamEntry *t = new TeamEntry();
	t->Freq = Freq;
	memcpy(t->Owner, Owner, 20);
	t->Password = Password;

	Box.append(t);

	return t;
}

Uint32 TeamBox::CountOwned(char *Owner)
{
	Uint32 ret = 0;

	_listnode <TeamEntry> *parse = Box.head;
	while (parse)
	{
		if (CMPSTR(parse->item->Owner, Owner))
			ret++;

		parse = parse->next;
	}

	return ret;
}

TeamInviteEntry *TeamBox::AddInvite(Uint16 Freq, char *Name)
{
	RemoveInvite(FindInvite(Freq, Name));

	TeamInviteEntry *t = new TeamInviteEntry;
	t->Freq = Freq;
	ZeroMemory(t->Name, 20);
	memcpy(t->Name, Name, min(STRLEN(Name), 19));
	t->Timestamp = GetTickCount();
	InviteBox.append(t);

	return t;
}

TeamInviteEntry *TeamBox::FindInvite(Uint16 Freq, char *Name)
{
	_listnode <TeamInviteEntry> *parse = InviteBox.head;
	while (parse)
	{
		if (parse->item->Freq == Freq && CMPSTR(parse->item->Name, Name))
			return parse->item;

		parse = parse->next;
	}

	return NULL;
}

void TeamBox::ExpireInvites(Uint32 Seconds)
{
	register Uint32 i = GetTickCount();

	_listnode <TeamInviteEntry> *parse = InviteBox.head;
	while (parse)
	{
		if (parse->item->Timestamp + (Seconds * 1000) < i)
		{
			if (parse->next)
			{
				parse = parse->next;
				InviteBox.kill(parse->last);
				continue;
			}
			else
			{
				InviteBox.kill(parse);
				break;
			}
		}

		parse = parse->next;
	}
}

//////////////////////////////

String botInfo::getDateString(tm *CDate)
{
	String s;

	if (CDate)
	{
		s += CDate->tm_mday + 1;
		if (CDate->tm_mon + 1 < 10)
			s += ".0";
		else
			s += ".";
		s += CDate->tm_mon + 1;
		s += ".";
		s += CDate->tm_year + 1900;
	}
	else
	{
		s += "(INVALID DATE)";
	}
	return s;
}

String botInfo::getTimeString(tm *CDate)
{
	String s;

	if (CDate)
	{
		s += CDate->tm_hour;
		if (CDate->tm_min < 10)
			s += ":0";
		else
			s += ":";
		s += CDate->tm_min;
		if (CDate->tm_sec < 10)
			s += ":0";
		else
			s += ":";
		s += CDate->tm_sec;
	}
	else
	{
		s += "(INVALID TIME)";
	}
	return s;
}

Uint32 botInfo::strFirstInstanceOf(char *msg, char *msg2)
{
	Uint32 len1 = STRLEN(msg), len2 = STRLEN(msg2);

	if (len1 < len2)
		return -1;

	for (Uint32 i = 0; i <= (len1 - len2); i++)
	{
		if (msg[i] == msg2[0])
		{
			for (Uint32 i2 = 0; i2 <= len2; i2++)
			{
				if (msg2[i2] == 0)
					return i;

				if (msg[i + i2] != msg2[i2])
					break;
			}
		}
	}

	return -1;
}

String botInfo::strReplace(char *msg, char *from, char *to)
{
	String ret;

	Uint32 fromlen = STRLEN(from);

	Uint32 pos;

	while (true)
	{
		pos = strFirstInstanceOf(msg, from);
		if (pos == -1)
		{
			/// Append the tail and go ///

			ret += msg;
			break;
		}

		ret.append(msg, pos);
		ret += to;
		msg += pos + fromlen;
	}

	return ret;
}

void botInfo::LoadSettings()
{
	char Path[520];
	GetCurrentDirectory(520, Path);
	strcat(Path, "\\freqs.ini");

	char buffer[16];

	ZeroMemory(buffer, 16);
	GetPrivateProfileString("Config", "MaxOwnedFreqs", "3", buffer, 16, Path);
	if (!isNumeric(&buffer[0]))
	{
		tell(makeEcho("freqs.ini error: Config:MaxOwnedFreqs must be numeric"));
		MaxOwnedFreqs = 3;
	}
	else
		MaxOwnedFreqs = getInteger(&buffer[0], 10);

	ZeroMemory(buffer, 16);
	GetPrivateProfileString("Config", "InviteExpireSeconds", "30", buffer, 16, Path);
	if (!isNumeric(&buffer[0]))
	{
		tell(makeEcho("freqs.ini error: Config:InviteExpireSeconds must be numeric"));
		InviteExpireSeconds = 30;
	}
	else
		InviteExpireSeconds = getInteger(&buffer[0], 10);

	ZeroMemory(buffer, 16);
	GetPrivateProfileString("Config", "AutoSaveDelay", "300", buffer, 16, Path);
	if (!isNumeric(&buffer[0]))
	{
		tell(makeEcho("freqs.ini error: Config: must be numeric"));
		AutoSaveDelay = 300;
	}
	else
		AutoSaveDelay = getInteger(&buffer[0], 10);
}

//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}

//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];

			if (!countdown[0])
			{
				TBox.ExpireInvites(InviteExpireSeconds);

				countdown[0] = VEX_INVITE_EXPIRE_DELAY;
			}

			if (!countdown[1])
			{
				tell(makeEcho("AutoSave..."));
				TBox.SaveToDisk();

				countdown[1] = AutoSaveDelay;
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
			bool billeronline = event.p[2] ? true : false;

			SpecFreq = me->team;
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_BallMove:
		{
			PBall *ball = (PBall*)event.p[0];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];
			weaponInfo wi;
			wi.n = *(Uint16*)&event.p[1];
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
			int oldship = (int)event.p[1];
			int oldteam = (int)event.p[2];
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];			
			int oldteam = (int)event.p[1];
			int oldship = (int)event.p[2];
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
			int oldship = (int)event.p[2];

			TeamEntry *t = TBox.FindTeam(p->team);

			if (t)
			{
				if ((!CMPSTR(t->Owner, p->name) && !TBox.FindInvite(p->team, p->name)) && !(t->AllowedSquad[0] && CMPSTR(t->AllowedSquad, p->squad)))
				{
					sendPrivate(p, "You are not allowed on that freq");
					sendPrivate(p, "*setfreq 0");
				}
				else
				{
					time(&t->LastUsed);

					if (t->GreetMessage[0])
					{                        
						String s = t->GreetMessage;
						s = strReplace(s.msg, "%name%", p->name);
						s = strReplace(s.msg, "%fnum%", getString(p->team, 10, 0, false).msg);
						s = strReplace(s.msg, "%fname%", t->Name);
						s = strReplace(s.msg, "%fsquad%", t->AllowedSquad);
						s = strReplace(s.msg, "%sq%", p->squad);
						s = strReplace(s.msg, "%points%", getString((p->score.flagPoints + p->score.killPoints), 10, 0, false).msg);

						sendPrivate(p, s.msg);
					}
				}
			}
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
			case EVENT_SelfShipReset:
		{
		}
		break;
	case EVENT_SelfPrize:
		{
			Uint16 prize = (Uint16)event.p[0];
			Uint16 count = (Uint16)event.p[1];
		}
		break;
	case EVENT_SelfUFO:
		{
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				break;
			case MSG_PublicMacro:
			case MSG_Public:
			case MSG_Team:
			case MSG_TeamPrivate:
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
	case EVENT_PositionHook:
		{
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];
			map = (BYTE*)event.p[3];

			if (major > CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin is out of date."));

				CONNECTION_DENIED = true;

				return;
			}
			else if ((major < CORE_MAJOR_VERSION) || (minor < CORE_MINOR_VERSION))
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin requires the latest version of MERVBot."));

				CONNECTION_DENIED = true;

				return;
			}
			else
				tell(makeEcho("DLL plugin connected."));

			LoadSettings();

			countdown[1] = AutoSaveDelay;
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s.msg);
}
